from angorapy.analysis.investigators.base_investigator import Investigator
from angorapy.analysis.investigators.predictability import Predictability
# from angorapy.analysis.investigators.dynamics import Dynamics, LatentDynamics
from angorapy.analysis.investigators.correlation import Correlation
from angorapy.analysis.investigators.activity_autocorrelation import ActivityAutocorrelation